/*
 * HealthcardIO - a library providing functionality for accessing healthcards
 *
 * Copyright (C) 2017  gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.api;

import de.gematik.security.impl.HealthcardContext;
import de.gematik.security.iso7816.Iso7816UnsuccessfulExecutionException;

import javax.smartcardio.CardException;
import java.io.IOException;

/**
 * @author André Szdzuy
 */
public interface CryptoCommands {

    /**
     * Establishes secure messaging
     * <p/>
     * <p> The PaceInfo is read from the card. The card access number is taken from the card object.
     * After secure messaging is established the apdu exchange is secured using the algorithms
     * specified in the PaceInfo.
     *
     * @param can Card Access Number
     * @throws CardException if the secure messaging couldn't be established successfully
     */
    void secureChannel(String can) throws CardException;

    /**
     * Performs PSO Compute Digital Signature (RSA1.5) without message recovery.
     *
     * @param input data to sign
     * @param dfSpecificPrivateKeyRef keyIdentifier of a DF specific private key
     * @return signature
     *
     * @throws CardException any Card IO exception
     * @throws IOException any general IO exception
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *         (SW is stored in detailed message)
     */
    byte[] doRSASSA_PKCS1_V1_5_SIGN(byte[] input, int dfSpecificPrivateKeyRef) throws CardException, IOException, Iso7816UnsuccessfulExecutionException;

    /**
     * Performs PSO Compute Digital Signature without message recovery.
     *
     * @param input data to sign
     * @param dfSpecificPrivateKeyRef keyIdentifier of a DF specific private key
     * @return signature
     *
     * @throws CardException any Card IO exception
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *         (SW is stored in detailed message)
     */
    byte[] doPSSSigner(byte[] input, int dfSpecificPrivateKeyRef) throws CardException, Iso7816UnsuccessfulExecutionException;

    /**
     * Performs PSO Compute Digital Signature with message recovery.
     *
     * @param input data to sign
     * @param dfSpecificPrivateKeyRef keyIdentifier of a DF specific private key
     * @return signature
     *
     * @throws CardException any Card IO exception
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *         (SW is stored in detailed message)
     */
    byte[] doIso9796d2PSSSigner(byte[] input, int dfSpecificPrivateKeyRef) throws CardException, Iso7816UnsuccessfulExecutionException;

    /**
     * Performs a Card2Card authentication.
     *
     *  Use Case 1: Freischalten einer EGK (this) durch einen HBA (otherCard), z.B. zum Zugriff auf die geschützten Stammdaten
     *  Use Case 2: Echtheitsprüfung einer EGK (otherCard) durch einen HBA (this)
     * @param otherCard the card that should be authenticated against this card
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *         (SW is stored in detailed message)
     * @throws CardException any Card IO exception
     * @throws IOException any general IO exception
     */
    void authenticateOtherCard(HealthcardContext otherCard) throws Iso7816UnsuccessfulExecutionException, CardException, IOException;

}
