/**
 *  Copyright 2011, Tobias Senger
 *  
 *  This file is part of animamea.
 *
 *  Animamea is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Animamea is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License   
 *  along with animamea.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.gematik.security.crypto;

import org.spongycastle.crypto.*;
import org.spongycastle.crypto.engines.AESFastEngine;
import org.spongycastle.crypto.macs.CMac;
import org.spongycastle.crypto.modes.CBCBlockCipher;
import org.spongycastle.crypto.paddings.ISO7816d4Padding;
import org.spongycastle.crypto.paddings.PaddedBufferedBlockCipher;
import org.spongycastle.crypto.params.KeyParameter;
import org.spongycastle.crypto.params.ParametersWithIV;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

/**
 * This class provides cryptographic operations on the base of the Bouncycastle library.
 *
 * @author Tobias Senger (tobias@t-senger.de)
 */

public class AmAESCrypto extends AmCryptoProvider {

	private byte[] keyBytes = null;
	private KeyParameter keyP = null;
	private byte[] IV = null;
	private byte[] sscBytes = null;

	public static int BLOCK_SIZE = 16;

	private void initCiphers(byte[] key, byte[] iv) {

		// get the keyBytes
		keyBytes = new byte[key.length];
		System.arraycopy(key, 0, keyBytes, 0, key.length);

		keyP = new KeyParameter(keyBytes);

		// get the IV
		IV = new byte[BLOCK_SIZE];
		System.arraycopy(iv, 0, IV, 0, IV.length);

		// create the ciphers
		// AES block cipher in CBC mode with ISO7816d4 padding
		encryptCipher = new PaddedBufferedBlockCipher(new CBCBlockCipher(
				new AESFastEngine()), new ISO7816d4Padding());

		decryptCipher = new PaddedBufferedBlockCipher(new CBCBlockCipher(
				new AESFastEngine()), new ISO7816d4Padding());

		// create the IV parameter
		ParametersWithIV parameterIV = new ParametersWithIV(keyP, IV);

		encryptCipher.init(true, parameterIV);
		decryptCipher.init(false, parameterIV);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.tsenger.animamea.crypto.AmCryptoProvider#init(byte[], byte[])
	 */
	@Override
	public void init(byte[] keyBytes, byte[] ssc) {

		sscBytes = ssc.clone();

		byte[] iv = encryptBlock(keyBytes, sscBytes);

		initCiphers(keyBytes, iv);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.tsenger.animamea.crypto.AmCryptoProvider#getMAC(byte[])
	 */
	@Override
	public byte[] getMAC(byte[] data) {

		byte[] n = new byte[sscBytes.length + data.length];
		System.arraycopy(sscBytes, 0, n, 0, sscBytes.length);
		System.arraycopy(data, 0, n, sscBytes.length, data.length);
		n = addPadding(n);

		BlockCipher cipher = new AESFastEngine();
		Mac mac = new CMac(cipher, 64);

		mac.init(keyP);
		mac.update(n, 0, n.length);
		byte[] out = new byte[mac.getMacSize()];

		mac.doFinal(out, 0);

		return out;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.tsenger.animamea.crypto.AmCryptoProvider#getMAC(byte[], byte[])
	 */
	@Override
	public byte[] getMAC(byte[] key, byte[] data, boolean padding) {
		if (padding) {
			data = addPadding(data);
		}

		BlockCipher cipher = new AESFastEngine();
		Mac mac = new CMac(cipher, 64);

		KeyParameter keyP = new KeyParameter(key);
		mac.init(keyP);

		mac.update(data, 0, data.length);

		byte[] out = new byte[mac.getMacSize()];

		mac.doFinal(out, 0);

		return out;
	}

	@Override
	public byte[] encryptBlock(byte[] key, byte[] z) {
		return cryptBlock(key, z, true);
	}

	@Override
	public byte[] decryptBlock(byte[] key, byte[] z) {
		return cryptBlock(key, z, false);
	}

	private byte[] cryptBlock(byte[] key, byte[] z, boolean forEncryption) {
		byte[] s = new byte[BLOCK_SIZE];
		KeyParameter encKey = new KeyParameter(key);
		BlockCipher cipher = new AESFastEngine();
		cipher.init(forEncryption, encKey);
		cipher.processBlock(z, 0, s, 0);
		return s;
	}

	@Override
	public byte[] encrypt(byte[] key, byte[] data) throws AmCryptoException {
		return crypt(key, data, true);
	}

	@Override
	public byte[] decrypt(byte[] key, byte[] data) throws AmCryptoException {
		return crypt(key, data, false);
	}

	private byte[] crypt(byte[] key, byte[] data, boolean forEncryption) throws AmCryptoException {
		KeyParameter keyP = new KeyParameter(key);
		byte[] iv = new byte[BLOCK_SIZE];	// IV filled with zeros
		ParametersWithIV parameterIV = new ParametersWithIV(keyP, iv);
		BufferedBlockCipher cipher = new BufferedBlockCipher(new CBCBlockCipher(new AESFastEngine()));
		cipher.init(forEncryption, parameterIV);

		ByteArrayInputStream bin = new ByteArrayInputStream(data);
		ByteArrayOutputStream bout = new ByteArrayOutputStream();

		int noBytesRead = 0; // number of bytes read from input
		int noBytesProcessed = 0; // number of bytes processed

		try {
			while ((noBytesRead = bin.read(buf)) >= 0) {
				noBytesProcessed = cipher.processBytes(buf, 0, noBytesRead,
						obuf, 0);
				bout.write(obuf, 0, noBytesProcessed);
			}
		} catch (DataLengthException e) {
			throw new AmCryptoException(e);
		} catch (IllegalStateException e) {
			throw new AmCryptoException(e);
		} catch (IOException e) {
			throw new AmCryptoException(e);
		}

		try {
			noBytesProcessed = cipher.doFinal(obuf, 0);
			bout.write(obuf, 0, noBytesProcessed);

			bout.flush();

			bin.close();
			bout.close();
			return bout.toByteArray();
		} catch (DataLengthException e) {
			throw new AmCryptoException(e);
		} catch (IllegalStateException e) {
			throw new AmCryptoException(e);
		} catch (InvalidCipherTextException e) {
			throw new AmCryptoException(e);
		} catch (IOException e) {
			throw new AmCryptoException(e);
		}
	}

	@Override
	public int getBlockSize() {
		return BLOCK_SIZE;
	}

}
