/*
 * HealthcardIO - a library providing functionality for accessing healthcards
 *
 * Copyright (C) 2017  gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.impl;

import de.gematik.security.Utilities.CardFileSystem;
import de.gematik.security.Utilities.CardItemFile;
import de.gematik.security.Utilities.Utilities;
import de.gematik.security.api.CryptoCommands;
import de.gematik.security.api.GematikCommands;
import de.gematik.security.api.Iso7816Commands;
import de.gematik.security.crypto.AmAESCrypto;
import de.gematik.security.crypto.AmCryptoProvider;
import de.gematik.security.iso7816.Constants;
import de.gematik.security.iso7816.FCP;
import de.gematik.security.iso7816.Iso7816UnsuccessfulExecutionException;
import de.gematik.security.iso7816.TlvData;
import de.gematik.security.iso9564.PinBlock;
import de.gematik.security.iso9564.PinBlockException;
import de.gematik.security.tr03110.*;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.json.JSONException;
import org.spongycastle.asn1.*;
import org.spongycastle.crypto.Digest;
import org.spongycastle.crypto.digests.SHA256Digest;
import org.spongycastle.jce.ECNamedCurveTable;
import org.spongycastle.math.ec.ECCurve;
import org.spongycastle.math.ec.ECPoint;

import javax.smartcardio.CardChannel;
import javax.smartcardio.CardException;
import javax.smartcardio.CommandAPDU;
import javax.smartcardio.ResponseAPDU;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Properties;

/**
 * The central class of the Healthcard IO library. Provides implementations of all supported
 * card operations. Can be used with any SmartcardIO implementation, e.g. SmartcardioNFC or Sun's
 * PC/SC which is part of the JRE. The SmartcardIO implementation is plugged in with the constructor.
 *
 * @author Ronald Koenig
 * @author André Szdzuy
 */
public class HealthcardContext implements Iso7816Commands, CryptoCommands, GematikCommands {
    private static final Log LOG = LogFactory.getLog(HealthcardContext.class);

    private static final int PRIVATE_KEY_MODULUS_LENGTH = 2048;

    private static final int SK_CAN = 2; //identifier of secret key "card access number"
    private CardChannel channel;
    private SecureMessaging secureMessaging = null;
    private boolean isExtendedLengthApduSupported = false;
    private int maxLength = 256;

    /**
     * Constructs a new HealthcardContext.
     *
     * @param channel the SmartcardIO CardChannel instance on which to perform the cards commands,
     *                this way different SmartcardIO implementations can be plugged in
     * @param isExtendedLengthApduSupported isExtendedLengthApduSupported?
     * @throws CardException if an error occurred
     */
    public HealthcardContext(CardChannel channel, boolean isExtendedLengthApduSupported) throws CardException {
        this.channel = channel;
        if(isExtendedLengthApduSupported){
            this.isExtendedLengthApduSupported = true;
            this.maxLength = 65536;
        }
    }

    /**
     * @return the SmartcardIO CardChannel instance
     */
    public CardChannel getChannel() {
        return channel;
    }

    /**
     * @return isExtendedLengthApduSupported?
     */
    public boolean isExtendedLengthApduSupported() {
        return isExtendedLengthApduSupported;
    }

    /**
     * @return max length
     */
    public int getMaxLength() {
        return maxLength;
    }

    /**
     * Set max length.
     *
     * @param maxLength max length
     */
    public void setMaxLength(int maxLength) {
        this.maxLength = maxLength;
    }

    /**
     * Close this HealthcardContext.
     */
    public void close() {
        disableSecureMessaging();
    }

    /**
     * Disable SecureMessaging.
     */
    public void disableSecureMessaging() {
        // TODO find a better way to pass SecureMessaging instance between classes HealthcardContext and NfcCardChannel
        // (the problem is that HealthcardContext knows only CardChannel and not NfcCardChannel)
        Properties props = System.getProperties();
        props.remove("secureMessaging");
        System.setProperties(props);
    }

    /**
     * Enable SecureMessaging.
     */
    public void enableSecureMessaging() {
        // TODO find a better way to pass SecureMessaging instance between classes HealthcardContext and NfcCardChannel
        // (the problem is that HealthcardContext knows only CardChannel and not NfcCardChannel)
        if (secureMessaging != null) {
            Properties props = System.getProperties();
            props.put("secureMessaging", secureMessaging);
            System.setProperties(props);
        }
    }

    /**
     * Adjust channel. The given CLA byte is "enriched" with the current channel number.
     *
     * @param cla CLA byte
     * @return adjusted CLA byte
     * @throws CardException if an error occurred
     */
    public byte adjustChannel(byte cla) throws CardException {
        return Utilities.adjustChannel(channel.getChannelNumber(), cla);
    }

    /*-------------------------- Methods that implement interface CryptoCommands --------------------------*/

    @Override
    public void secureChannel(String can) throws CardException {
        try {

            //Select EF_MF - "00 A4 00 0C 02 3F 00"
            ResponseAPDU responseAPDU = channel.transmit(new CommandAPDU(0x00, 0xA4, 0x00, 0x0C, new byte[]{(byte) 0x3f, 0x00}));
            assertResponseStatus(responseAPDU, Constants.APDU_RESPONSE_OK);

            //Select EF_CARDACCESS - "00 A4 02 0C 02 01 1C"
            responseAPDU = channel.transmit(new CommandAPDU(0x00, 0xA4, 0x02, 0x0C, new byte[]{(byte) 0x01, 0x1C}));
            assertResponseStatus(responseAPDU, Constants.APDU_RESPONSE_OK);

            //Read EF_CARDACCESS - "00 B0 00 00 80"
            responseAPDU = channel.transmit(new CommandAPDU(0x00, 0xB0, 0x00, 0x00, 0x80));
            assertResponseStatus(responseAPDU, Constants.APDU_RESPONSE_TOO_MUCH_DATA);
            ByteArrayInputStream bis = new ByteArrayInputStream(responseAPDU.getData());
            ASN1InputStream ais = new ASN1InputStream(bis);
            PaceInfo paceInfo = new PaceInfo((DLSequence) ((DLSet) ais.readObject()).getObjectAt(0));

            //MSE:AT - "00 22 C1 A4 ..."
            TlvData apduData = new TlvData().putData(0x80, paceInfo.getProtocolBytes()).putData(0x83, new byte[]{SK_CAN});
            responseAPDU = channel.transmit(new CommandAPDU(0x00, 0x22, 0xC1, 0xA4, apduData.getBytes()));
            assertResponseStatus(responseAPDU, Constants.APDU_RESPONSE_OK);

            //GA:Get Nonce - "10 86 00 00 ..."
            apduData = new TlvData().putData(0x7C, new byte[]{});
            responseAPDU = channel.transmit(new CommandAPDU(0x10, 0x86, 0x00, 0x00, apduData.getBytes(), 256));
            assertResponseStatus(responseAPDU, Constants.APDU_RESPONSE_OK);
            byte[] nonceZ = new TlvData(new TlvData(responseAPDU.getData()).getData(0x7c)).getData(0x80);

            //Calculate nonce_s
            byte[] key = KeyDerivationFunction.getAES128Key(can.getBytes(), KeyDerivationFunction.Mode.PASSWORD);
            AmCryptoProvider crypto = new AmAESCrypto();
            byte[] nonceS = crypto.decryptBlock(key, nonceZ);

            //getX1
            PaceECDH paceECDH = new PaceECDH(ECNamedCurveTable.getParameterSpec(paceInfo.getParameterIDString()));
            byte[] X1 = paceECDH.getX1(nonceS);

            //GA:Map Nonce - "10 86 00 00 ..."
            apduData = new TlvData().putData(0x7C, new TlvData().putData(0x81, X1));
            responseAPDU = channel.transmit(new CommandAPDU(0x10, 0x86, 0x00, 0x00, apduData.getBytes(), 256));
            assertResponseStatus(responseAPDU, Constants.APDU_RESPONSE_OK);
            byte[] Y1 = new TlvData(new TlvData(responseAPDU.getData()).getData(0x7c)).getData(0x82);

            //getX2
            byte[] X2 = paceECDH.getX2(Y1);

            //GA:Perform key agreement - ""10 86 00 00 ..."
            apduData = new TlvData().putData(0x7C, new TlvData().putData(0x83, X2));
            responseAPDU = channel.transmit(new CommandAPDU(0x10, 0x86, 0x00, 0x00, apduData.getBytes(), 256));
            assertResponseStatus(responseAPDU, Constants.APDU_RESPONSE_OK);
            byte[] Y2 = new TlvData(new TlvData(responseAPDU.getData()).getData(0x7c)).getData(0x84);

            //calculate ec point y
            byte[] S = paceECDH.getSharedSecret_K(Y2);
            byte[] kenc = KeyDerivationFunction.getAES128Key(S, KeyDerivationFunction.Mode.ENC);
            byte[] kmac = KeyDerivationFunction.getAES128Key(S, KeyDerivationFunction.Mode.MAC);


            ECCurve.Fp curve = (ECCurve.Fp) ECNamedCurveTable.getParameterSpec(paceInfo.getParameterIDString()).getCurve();
            ECPoint pointY = Converter.byteArrayToECPoint(Y2, curve);

            //create asn.1 auth token pcd
            ASN1EncodableVector asn1EncodableVector = new ASN1EncodableVector();
            asn1EncodableVector.add(new ASN1ObjectIdentifier(paceInfo.getProtocolID()));
            asn1EncodableVector.add(new DERTaggedObject(false, 6, new DEROctetString(pointY.getEncoded(false))));
            byte[] authToken = new DERApplicationSpecific(0x49, asn1EncodableVector).getEncoded();

            //calculate mac pcd
            byte[] macPCD = crypto.getMAC(kmac, authToken, false);

            //GA: Mutual Authentication - "00 86 00 00 ..."
            apduData = new TlvData().putData(0x7C, new TlvData().putData(0x85, macPCD));
            responseAPDU = channel.transmit(new CommandAPDU(0x00, 0x86, 0x00, 0x00, apduData.getBytes(), 256));
            assertResponseStatus(responseAPDU, Constants.APDU_RESPONSE_OK);
            byte[] macPICC = new TlvData(new TlvData(responseAPDU.getData()).getData(0x7c)).getData(0x86);

            //calculate ec point x
            ECPoint pointX = Converter.byteArrayToECPoint(X2, curve);

            //create asn.1 auth token icc
            asn1EncodableVector = new ASN1EncodableVector();
            asn1EncodableVector.add(new ASN1ObjectIdentifier(paceInfo.getProtocolID()));
            asn1EncodableVector.add(new DERTaggedObject(false, 6, new DEROctetString(pointX.getEncoded(false))));
            authToken = new DERApplicationSpecific(0x49, asn1EncodableVector).getEncoded();

            //calculate mac picc
            byte[] macPICC2 = crypto.getMAC(kmac, authToken, false);

            //compare calculated and received mac picc
            if (!Arrays.equals(macPICC, macPICC2)) {
                throw new CardException("MAC doesn't match");
            }
            secureMessaging = new SecureMessaging(crypto, kenc, kmac, new byte[crypto.getBlockSize()]);

            // TODO find a better way to pass SecureMessaging instance between classes HealthcardContext and NfcCardChannel
            // (the problem is that HealthcardContext knows only CardChannel and not NfcCardChannel)
            Properties props = System.getProperties();
            props.put("secureMessaging", secureMessaging);
            System.setProperties(props);
        } catch (IOException e) {
            throw new CardException("IOException");
        }
    }

    /**
     * Assert response status
     * <p/>
     *
     * @param responseAPDU response APDU to be checked
     * @param expectedSW   response status (status word)
     * @throws CardException if received response status doesn't match expected response status
     */
    public void assertResponseStatus(ResponseAPDU responseAPDU, int expectedSW) throws CardException {
        if (responseAPDU.getSW() != expectedSW) {
            throw new CardException(String.format("expected status: %#x but was: %#x", expectedSW, responseAPDU.getSW()));
        }
    }

    @Override
    public byte[] doRSASSA_PKCS1_V1_5_SIGN(byte[] input, int dfSpecificPrivateKeyRef) throws CardException, IOException, Iso7816UnsuccessfulExecutionException {
        // RSASSA_PKCS1_V1_5_SIGN with SHA256 (see PKCS#1V2.2 chapter 8.2.1)
        LOG.info("doRSASSA_PKCS1_V1_5_SIGN()");
        LOG.info("input: " + Utilities.ByteArrayToSpacedHexString(input));

        selectADF("A000000167455349474E");  // DF.ESIGN

        //generate signature
        //generate hash
        Digest digest = new SHA256Digest();
        int hLen = digest.getDigestSize();
        byte[] hashOfMessage = new byte[hLen];
        digest.update(input, 0, input.length);
        digest.doFinal(hashOfMessage, 0);

        //build digest info
        ASN1EncodableVector algorithmIdentifierVector = new ASN1EncodableVector();
        algorithmIdentifierVector.add(new ASN1ObjectIdentifier("2.16.840.1.101.3.4.2.1")); //SHA256
        algorithmIdentifierVector.add(DERNull.INSTANCE);
        ASN1EncodableVector digestInfoVector = new ASN1EncodableVector();
        digestInfoVector.add(new DERSequence(algorithmIdentifierVector));
        digestInfoVector.add(new DEROctetString(hashOfMessage)); //hash
        byte[] digestInfo = new DERSequence(digestInfoVector).getEncoded();

        // MSE AT
        String keyRef = Integer.toHexString(0x80 | dfSpecificPrivateKeyRef);
        String algId = "02";
        manageSE("84 01", keyRef, algId, 0x41, 0xB6);

        // PSO Compute Digital Signature without message recovery
        ResponseAPDU responseAPDU = channel.transmit(new CommandAPDU(0x00, 0x2A, 0x9E, 0x9A, digestInfo, 256));
        if (responseAPDU.getSW() != Constants.APDU_RESPONSE_OK) {
            throw new CardException("RSASSA_PKCS1_V1_5 - unsuccessful signature generation");
        }
        byte[] signature = responseAPDU.getData();

        return signature;
    }

    @Override
    public byte[] doPSSSigner(byte[] input, int dfSpecificPrivateKeyRef) throws CardException, Iso7816UnsuccessfulExecutionException {
        LOG.info("doPSSSigner()");
        LOG.info("input: " + Utilities.ByteArrayToSpacedHexString(input));

        selectADF("A000000167455349474E");  // DF.ESIGN

        Digest digest = new SHA256Digest();
        int hLen = digest.getDigestSize();
        byte[] hashOfMessage = new byte[hLen];
        digest.update(input, 0, input.length);
        digest.doFinal(hashOfMessage, 0);

        // MSE AT
        String keyRef = Integer.toHexString(0x80 | dfSpecificPrivateKeyRef);
        String algId = "05";
        manageSE("84 01", keyRef, algId, 0x41, 0xB6);

        // PSO Compute Digital Signature without message recovery
        ResponseAPDU responseAPDU = channel.transmit(new CommandAPDU(0x00, 0x2A, 0x9E, 0x9A, hashOfMessage, 256));
        if (responseAPDU.getSW() != Constants.APDU_RESPONSE_OK) {
            throw new CardException("PSSSigner - unsuccessful signature generation");
        }
        byte[] signature = responseAPDU.getData();

        return signature;
    }

    @Override
    public byte[] doIso9796d2PSSSigner(byte[] input, int dfSpecificPrivateKeyRef) throws CardException, Iso7816UnsuccessfulExecutionException {
        LOG.info("doIso9796d2PSSSigner()");
        LOG.info("input: " + Utilities.ByteArrayToSpacedHexString(input));

        selectADF("A000000167455349474E");  // DF.ESIGN

        Digest digest = new SHA256Digest();
        int c = (PRIVATE_KEY_MODULUS_LENGTH - digest.getDigestSize() * 8 * 2 - 8 - 2) / 8;
        byte[] hashOfNonRecoverablePart = new byte[digest.getDigestSize()];
        byte[] recoverablePart;
        if (input.length > c) {
            digest.update(input, c, input.length - c);
            recoverablePart = Arrays.copyOfRange(input, 0, c);
        } else {
            recoverablePart = input.clone();
        }
        digest.doFinal(hashOfNonRecoverablePart, 0);
        ASN1EncodableVector vector = new ASN1EncodableVector();
        vector.add(new DLTaggedObject(false, 0x0, new DEROctetString(recoverablePart)));
        vector.add(new DLTaggedObject(false, 0x10, new DEROctetString(hashOfNonRecoverablePart)));
        byte[] inputParam = new DERApplicationSpecific(0, vector).getContents();

        // MSE AT
        String keyRef = Integer.toHexString(0x80 | dfSpecificPrivateKeyRef);
        String algId = "07";
        manageSE("84 01", keyRef, algId, 0x41, 0xB6);

        // PSO Compute Digital Signature with message recovery
        ResponseAPDU responseAPDU = channel.transmit(new CommandAPDU(0x00, 0x2A, 0x9E, 0xAC, inputParam, 256));
        if (responseAPDU.getSW() != Constants.APDU_RESPONSE_OK) {
            throw new CardException("Iso9796d2PSSSigner - unsuccessful signature generation");
        }

        byte[] signature = responseAPDU.getData();

        return signature;
    }

    // Use Case 1: Freischalten einer EGK (this) durch einen HBA (otherCard), z.B. zum Zugriff auf die geschützten Stammdaten
    // Use Case 2: Echtheitsprüfung einer EGK (otherCard) durch einen HBA (this)
    // die Kommentare und Variablennamen innerhalb der Methode beziehen sich zum besseren Verständnis auf Use Case 1
    @Override
    public void authenticateOtherCard(HealthcardContext otherCard) throws Iso7816UnsuccessfulExecutionException, CardException, IOException {
        // 1. Freischaltung eGK:

        // verify HBA's CA cert on EGK
        this.verifyCertificate(otherCard, "2F07");

        // verify HBA's AUT cert on EGK
        byte[] certHBA = this.verifyCertificate(otherCard, "2F06");

        // load ICCSN from file EF.GDO -> HBA
        byte[] iccsnHBA = otherCard.readICCSN();
        LOG.info("iccsnHBA from EF.GDO: " + Utilities.ByteArrayToSpacedHexString(iccsnHBA));
        // laut Silvias Dokument soll die iccsnHBA aus EF.GDO mit der aus dem HBA-Aut-Zertifikat
        // übereinstimmen, tut sie aber nicht
        // deshalb muss die iccsnHBA direkt aus dem HBA-Aut-Zertifikat extrahiert werden
        iccsnHBA = new byte[12];
        System.arraycopy(certHBA, 103, iccsnHBA, 0, 12);
        LOG.info("iccsnHBA from certHBA: " + Utilities.ByteArrayToSpacedHexString(iccsnHBA));

        // load ICCSN from file EF.GDO -> EGK
        byte[] iccsnEGK = this.readICCSN();
        LOG.info("iccsnEGK: " + Utilities.ByteArrayToSpacedHexString(iccsnEGK));

        // MSE SET -> HBA (select key for INTERNAL AUTHENTICATE)
        otherCard.selectADF();
        int identifier;
        if (otherCard.isEGK()) {
            identifier = 0x09;
        } else {
            identifier = 0x06;
        }
        String keyRef = Utilities.toHexString(0x00 | identifier);   // location = 0x00, identifier = 0x06
        String algId = "00";    // elcRoleAuthentication (MF/PrK.HPC.AUTR_CVC.E256)
        otherCard.manageSE("84 01", keyRef, algId, 0x41, 0xA4);

        // MSE SET -> EGK (select key for EXTERNAL AUTHENTICATE)
        this.selectADF();
        keyRef = Utilities.ByteArrayToSpacedHexString(iccsnHBA);
        algId = "00";    // elcRoleCheck
        this.manageSE("83 0C", keyRef, algId, 0x81, 0xA4);

        // GET CHALLENGE -> EGK
        ResponseAPDU responseAPDU = this.getChannel().transmit(new CommandAPDU(0x00, 0x84, 0x00, 0x00, 0x10));
        this.assertResponseStatus(responseAPDU, Constants.APDU_RESPONSE_OK);
        byte[] rnd1 = responseAPDU.getData();

        // construct input for INTERNAL AUTHENTICATE from rnd1 and iccsnEGK
        // iccsnEGK has 12 bytes, but we only append its last 8 bytes to rnd1
        byte[] iaInput = new byte[rnd1.length + 8];
        System.arraycopy(rnd1, 0, iaInput, 0, rnd1.length);
        System.arraycopy(iccsnEGK, 4, iaInput, rnd1.length, 8);

        // INTERNAL AUTHENTICATE -> HBA
        responseAPDU = otherCard.getChannel().transmit(new CommandAPDU(0x00, 0x88, 0x00, 0x00, iaInput, 0x100));
        otherCard.assertResponseStatus(responseAPDU, Constants.APDU_RESPONSE_OK);
        byte[] iaOutput = responseAPDU.getData();

        // EXTERNAL AUTHENTICATE -> EGK
        responseAPDU = this.getChannel().transmit(new CommandAPDU(0x00, 0x82, 0x00, 0x00, iaOutput));
        this.assertResponseStatus(responseAPDU, Constants.APDU_RESPONSE_OK);
    }

    private byte[] verifyCertificate(HealthcardContext otherCard, String fid) throws Iso7816UnsuccessfulExecutionException, CardException, IOException {
        // Zertifikat vom HBA laden
        otherCard.selectADF();
        otherCard.selectEF(fid);
        byte[] certData = otherCard.read();
        LOG.info("certData: " + Utilities.ByteArrayToSpacedHexString(certData));

        byte[] car = new byte[8];
        System.arraycopy(certData, 14, car, 0, 8);
        LOG.info("CAR: " + Utilities.ByteArrayToSpacedHexString(car));

        // MSE SET -> EGK (Auswahl des für das nächste Kommando zu verwendenden Schlüssels)
        this.selectADF();
        String keyRef = Utilities.ByteArrayToSpacedHexString(car);
        this.manageSE("83 08", keyRef, 0x81, 0xB6);

        // PSO Verify Certificate
        byte[] certWithoutPrefix = new byte[certData.length - 4];
        System.arraycopy(certData, 4, certWithoutPrefix, 0, certData.length - 4);
        LOG.info("certWithoutPrefix: " + Utilities.ByteArrayToSpacedHexString(certWithoutPrefix));

        ResponseAPDU responseAPDU = this.getChannel().transmit(new CommandAPDU(0x00, 0x2A, 0x00, 0xBE, certWithoutPrefix));
        this.assertResponseStatus(responseAPDU, Constants.APDU_RESPONSE_OK);

        return certData;
    }

    /*-------------------------- Methods that implement interface Iso7816Commands -------------------------*/

    @Override
    public void selectADF() throws Iso7816UnsuccessfulExecutionException, CardException {
        selectADF(false);
    }

    @Override
    public void selectADF(boolean nextOccurrence) throws Iso7816UnsuccessfulExecutionException, CardException {
        ResponseAPDU responseAPDU = channel.transmit(new CommandAPDU(0x00, 0xA4, 0x04, 0xC | (nextOccurrence ? 0x2 : 0x0)));
        assertSelectResponse(responseAPDU);
    }

    @Override
    public void selectADF(String applicationIdentifier) throws Iso7816UnsuccessfulExecutionException, CardException {
        selectADF(applicationIdentifier, false);
    }

    @Override
    public void selectADF(String applicationIdentifier, boolean nextOccurrence) throws Iso7816UnsuccessfulExecutionException, CardException {
        ResponseAPDU responseAPDU = channel.transmit(new CommandAPDU(0x00, 0xA4, 0x04, 0xC | (nextOccurrence ? 0x2 : 0x0), Utilities.HexStringToByteArray(applicationIdentifier)));
        assertSelectResponse(responseAPDU);
    }

    @Override
    public void selectDF() throws Iso7816UnsuccessfulExecutionException, CardException {
        ResponseAPDU responseAPDU = channel.transmit(new CommandAPDU(0x00, 0xA4, 0x03, 0xC));
        assertSelectResponse(responseAPDU);
    }

    @Override
    public void selectDF(String fileIdentifier) throws Iso7816UnsuccessfulExecutionException, CardException {
        selectDF(fileIdentifier, false);
    }

    @Override
    public void selectDF(String fileIdentifier, boolean nextOccurrence) throws Iso7816UnsuccessfulExecutionException, CardException {
        ResponseAPDU responseAPDU = channel.transmit(new CommandAPDU(0x00, 0xA4, 0x01, 0xC | (nextOccurrence ? 0x2 : 0x0), Utilities.HexStringToByteArray(fileIdentifier)));
        assertSelectResponse(responseAPDU);
    }

    @Override
    public void selectEF(String fileIdentifier) throws Iso7816UnsuccessfulExecutionException, CardException {
        ResponseAPDU responseAPDU = channel.transmit(new CommandAPDU(0x00, 0xA4, 0x02, 0xC, Utilities.HexStringToByteArray(fileIdentifier)));
        assertSelectResponse(responseAPDU);
    }

    @Override
    public FCP selectADFAndReturnFCP() throws Iso7816UnsuccessfulExecutionException, CardException {
        return selectADFAndReturnFCP(false);
    }

    @Override
    public FCP selectADFAndReturnFCP(boolean nextOccurrence) throws Iso7816UnsuccessfulExecutionException, CardException {
        ResponseAPDU responseAPDU = channel.transmit(new CommandAPDU(0x00, 0xA4, 0x04, 0x4 | (nextOccurrence ? 0x2 : 0x0), maxLength));
        assertSelectResponse(responseAPDU);
        return new FCP(responseAPDU);
    }

    @Override
    public FCP selectADFAndReturnFCP(String applicationIdentifier) throws Iso7816UnsuccessfulExecutionException, CardException {
        return selectADFAndReturnFCP(applicationIdentifier, false);
    }

    @Override
    public FCP selectADFAndReturnFCP(String applicationIdentifier, boolean nextOccurrence) throws Iso7816UnsuccessfulExecutionException, CardException {
        ResponseAPDU responseAPDU = channel.transmit(new CommandAPDU(0x00, 0xA4, 0x04, 0x4 | (nextOccurrence?0x2:0x0), Utilities.HexStringToByteArray(applicationIdentifier), maxLength));
        assertSelectResponse(responseAPDU);
        return new FCP(responseAPDU);
    }

    @Override
    public FCP selectDFAndReturnFCP() throws Iso7816UnsuccessfulExecutionException, CardException {
        ResponseAPDU responseAPDU = channel.transmit(new CommandAPDU(0x00, 0xA4, 0x03, 0x4, maxLength));
        assertSelectResponse(responseAPDU);
        return new FCP(responseAPDU);
    }

    @Override
    public FCP selectDFAndReturnFCP(String fileIdentifier) throws Iso7816UnsuccessfulExecutionException, CardException {
        return selectDFAndReturnFCP(fileIdentifier, false);
    }

    @Override
    public FCP selectDFAndReturnFCP(String fileIdentifier, boolean nextOccurrence) throws Iso7816UnsuccessfulExecutionException, CardException {
        ResponseAPDU responseAPDU = channel.transmit(new CommandAPDU(0x00, 0xA4, 0x01, 0x4 | (nextOccurrence?0x2:0x0), Utilities.HexStringToByteArray(fileIdentifier), maxLength));
        assertSelectResponse(responseAPDU);
        return new FCP(responseAPDU);
    }

    @Override
    public FCP selectEFAndReturnFCP(String fileIdentifier) throws Iso7816UnsuccessfulExecutionException, CardException {
        ResponseAPDU responseAPDU = channel.transmit(new CommandAPDU(0x00, 0xA4, 0x02, 0x4, Utilities.HexStringToByteArray(fileIdentifier), maxLength));
        assertSelectResponse(responseAPDU);
        return new FCP(responseAPDU);
    }

    private void assertSelectResponse(ResponseAPDU responseAPDU) throws Iso7816UnsuccessfulExecutionException{
        int sw = responseAPDU.getSW();
        if (sw != Constants.APDU_RESPONSE_OK && sw != Constants.APDU_RESPONSE_FILE_DEACTIVATED && sw != Constants.APDU_RESPONSE_FILE_TERMINATED){
            throw new Iso7816UnsuccessfulExecutionException(responseAPDU);
        }
    }

    @Override
    public byte[] read() throws Iso7816UnsuccessfulExecutionException, CardException {
        return read(0, Integer.MAX_VALUE);
    }

    @Override
    public byte[] read(int offset, int len) throws Iso7816UnsuccessfulExecutionException, CardException {
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        try {
            ResponseAPDU responseAPDU;
            do {
                responseAPDU = channel.transmit(new CommandAPDU(0x00, 0xB0, ((offset + bos.size()) & 0xff00) >> 8, (offset + bos.size()) & 0xff, maxLength));
                if(responseAPDU.getSW()==Constants.APDU_RESPONSE_OK){
                    bos.write(responseAPDU.getData());
                }else{
                    if(responseAPDU.getSW()==Constants.APDU_RESPONSE_TOO_MUCH_DATA || responseAPDU.getSW()==Constants.APDU_RESPONSE_OFFSET_TOO_BIG){
                        bos.write(responseAPDU.getData());
                        break;
                    }else{
                        LOG.error(String.format("expected status 0x9000, 0x6282 or 0x6B00, but was: %#x", responseAPDU.getSW()));
                        throw new Iso7816UnsuccessfulExecutionException(responseAPDU);
                    }
                }
            }while (bos.size()<len);
            if(bos.size()>len){
                byte[] data = new byte[len];
                System.arraycopy(bos.toByteArray(), 0, data, 0, len);
                return data;
            }else{
                return bos.toByteArray();
            }
        }catch(IOException e){
            throw new CardException("IOException - writing to bos");
        }finally{
            try {
                bos.close();
            }catch (IOException e){
                throw new CardException("IOException - Cannot close bos");
            }
        }
    }

    @Override
    public byte[] readRecord(int recordNumber) throws Iso7816UnsuccessfulExecutionException, CardException {
        ResponseAPDU responseAPDU = channel.transmit(new CommandAPDU(0x00, 0xB2, recordNumber, 0x4, maxLength));
        if(responseAPDU.getSW()!=Constants.APDU_RESPONSE_OK){
            throw new Iso7816UnsuccessfulExecutionException(responseAPDU);
        }
        return responseAPDU.getData();
    }

    @Override
    public ArrayList<byte[]> readAllRecords() throws Iso7816UnsuccessfulExecutionException, CardException {
        ArrayList<byte[]> data = new ArrayList<>();
        for (int j = 1; ; j++) {
            try {
                data.add(readRecord(j));
            } catch (Iso7816UnsuccessfulExecutionException e) {
                if (e.getSW() == Constants.APDU_RESPONSE_RECORD_NOT_FOUND) {
                    break;
                } else {
                    throw e;
                }
            }
        }
        return data;
    }

    @Override
    public CardItemFile readEF(FCP parent, String fileIdentifier) throws Iso7816UnsuccessfulExecutionException, CardException, IOException, JSONException {
        return readEF(CardFileSystem.getEGKInstance(), parent, fileIdentifier);
    }

    @Override
    public CardItemFile readEF(CardFileSystem cfs, FCP parent, String fileIdentifier) throws Iso7816UnsuccessfulExecutionException, CardException, IOException, JSONException {
        FCP fcp = selectEFAndReturnFCP(fileIdentifier);
        ArrayList<byte[]> data;
        if (fcp.getFileDescriptor().length < 5) {
            data = new ArrayList<>();
            data.add(read());
        } else {
            data = readAllRecords();
        }

        return new CardItemFile(cfs, fcp, parent, data);
    }

    @Override
    public void update(byte[] data) throws Iso7816UnsuccessfulExecutionException, CardException {
        int offset = 0;
        int remainingLength = data.length;
        byte[] block;
        int blockSize = 1024;   // blockSize = 2048 und größer liefert den Error 6700
        int blockLength;
        do {
            if (remainingLength < blockSize) {
                blockLength = remainingLength;
            } else {
                blockLength = blockSize;
            }
            block = new byte[blockLength];
            System.arraycopy(data, offset, block, 0, blockLength);

            updateBlock(offset, block);

            offset = offset + blockSize;
            remainingLength = remainingLength - blockSize;
        } while (remainingLength > 0);
    }

    private void updateBlock(int offset, byte[] block) throws Iso7816UnsuccessfulExecutionException, CardException {
        ResponseAPDU responseAPDU =
                channel.transmit(new CommandAPDU(0x00, 0xD6, (offset & 0xff00) >> 8, offset & 0xff, block));
        assertResponseStatus(responseAPDU, Constants.APDU_RESPONSE_OK);
    }

    @Override
    public void verify(String verificationData) throws Iso7816UnsuccessfulExecutionException, CardException, PinBlockException {
        if (isEGK()) {
            verify(2, new PinBlock(verificationData));  // PIN Home
        } else {    // HBA or SMC-B
            verify(1, new PinBlock(verificationData));  // PIN.CH or PIN.SMC
        }
    }

    @Override
    public void verify(int secretReference, String verificationData) throws Iso7816UnsuccessfulExecutionException, CardException, PinBlockException {
        verify(secretReference, new PinBlock(verificationData));
    }

    @Override
    public void verify(int secretReference, PinBlock verificationData) throws Iso7816UnsuccessfulExecutionException, CardException {
        ResponseAPDU responseAPDU = channel.transmit(new CommandAPDU(0x00, 0x20, 0x00, secretReference, verificationData.getBytes()));
        if(responseAPDU.getSW()!=Constants.APDU_RESPONSE_OK){
            throw new Iso7816UnsuccessfulExecutionException(responseAPDU);
        }
    }

    @Override
    public void unlock(String puk, String newSecret) throws Iso7816UnsuccessfulExecutionException, CardException, PinBlockException {
        if (isEGK()) {
            unlock(2, puk, newSecret);  // PIN Home
        } else {    // HBA or SMC-B
            unlock(1, puk, newSecret);  // PIN.CH or PIN.SMC
        }
    }

    @Override
    public void unlock(int secretReference, String puk, String newSecret) throws Iso7816UnsuccessfulExecutionException, CardException, PinBlockException {
        unlock(secretReference,
               puk == null ? null : new PinBlock(puk),
               newSecret == null ? null : new PinBlock(newSecret));
    }

    @Override
    public void unlock(int secretReference, PinBlock puk, PinBlock newSecret) throws Iso7816UnsuccessfulExecutionException, CardException {
        int p1;
        byte[] data;
        if (puk == null) {
            if (newSecret == null) {
                p1 = 0x03;
                data = null;
            } else {
                p1 = 0x02;
                data = newSecret.getBytes();
            }
        } else {
            if (newSecret == null) {
                p1 = 0x01;
                data = puk.getBytes();
            } else {
                p1 = 0x00;
                byte[] pukBytes = puk.getBytes();
                byte[] newSecretBytes = newSecret.getBytes();
                data = new byte[pukBytes.length + newSecretBytes.length];
                System.arraycopy(pukBytes, 0, data, 0, pukBytes.length);
                System.arraycopy(newSecretBytes, 0, data, pukBytes.length, newSecretBytes.length);
            }
        }

        ResponseAPDU responseAPDU = channel.transmit(new CommandAPDU(0x00, 0x2C, p1, secretReference, data));
        if(responseAPDU.getSW()!=Constants.APDU_RESPONSE_OK){
            throw new Iso7816UnsuccessfulExecutionException(responseAPDU);
        }
    }

    @Override
    public void manageSE(String prefix, String keyRef, String algId, int p1, int p2) throws Iso7816UnsuccessfulExecutionException, CardException {
        byte[] data;
        if (algId == null) {
            data = Utilities.HexStringToByteArray(prefix + " " + keyRef);
        } else {
            data = Utilities.HexStringToByteArray(prefix + " " + keyRef + " 80 01 " + algId);
        }
        CommandAPDU commandAPDU = new CommandAPDU(0x00, 0x22, p1, p2, data);
//        Utilities.logCommandAPDU(commandAPDU);
        ResponseAPDU responseAPDU = channel.transmit(commandAPDU);

        if (responseAPDU.getSW() != Constants.APDU_RESPONSE_OK) {
            LOG.error(String.format("cannot set private key, error code = %#x", responseAPDU.getSW()));
            throw new Iso7816UnsuccessfulExecutionException(responseAPDU);
        }
    }

    @Override
    public void manageSE(String prefix, String keyRef, int p1, int p2) throws Iso7816UnsuccessfulExecutionException, CardException {
        manageSE(prefix, keyRef, null, p1, p2);
    }

    /*-------------------------- Methods that implement interface GematikCommands --------------------------*/

    @Override
    public boolean isEGK() throws Iso7816UnsuccessfulExecutionException, CardException {
        //SELECT DF.HCA
        return isSpecificCard("D27600000102");
    }

    @Override
    public boolean isHBA() throws Iso7816UnsuccessfulExecutionException, CardException {
        //SELECT DF.HPA
        return isSpecificCard("D27600014602");
    }

    @Override
    public boolean isSMCB() throws Iso7816UnsuccessfulExecutionException, CardException {
        //SELECT DF.SMA
        return isSpecificCard("D27600014607");
    }

    private boolean isSpecificCard(String aid) throws Iso7816UnsuccessfulExecutionException, CardException {
        try {
            selectADF(aid);
        } catch (Iso7816UnsuccessfulExecutionException e) {
            if (e.getSW() == Constants.APDU_RESPONSE_FILE_NOT_FOUND) {
                return false;
            } else {
                throw e;
            }
        }
        return true;
    }

    @Override
    public String getType() throws Iso7816UnsuccessfulExecutionException, CardException {
        if (isEGK()) {
            return "EGK";
        } else if (isHBA()) {
            return "HBA";
        } else if (isSMCB()) {
            return "SMC-B";
        } else {
            throw new CardException("Inserted card is neither EGK, HBA nor SMC-B.");
        }
    }

    @Override
    public byte[] readICCSN() throws Iso7816UnsuccessfulExecutionException, CardException {
        // load ICCSN from file EF.GDO
        selectADF();
        selectEF("2F02");
        return read();
    }

}