/*
 * HealthcardIO - a library providing functionality for accessing healthcards
 *
 * Copyright (C) 2017  gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.iso7816;

import de.gematik.security.Utilities.Utilities;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.spongycastle.asn1.*;

import javax.smartcardio.CardException;
import javax.smartcardio.ResponseAPDU;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.util.Enumeration;

/**
 * This class bundles some metadata for a file (DF or EF) on the card.
 * Should only be used in contexts where Extended Length Apdus are supported!
 *
 * @author Ronald Koenig
 * @author André Szdzuy
 */
public class FCP {
    private static final Log LOG = LogFactory.getLog(FCP.class);

    public static final int LCS_CREATION_STATE = 1;
    public static final int LCS_INITIALISATION_STATE = 3;
    public static final int LCS_OPERATIONAL_STATE_ACTIVATED = 5;
    public static final int LCS_OPERATIONAL_STATE_DEACTIVATED = 4;
    public static final int LCS_TERMINATION_STATE = 12;

    private byte[] fileDescriptor;
    private byte[] fileIdentifier;
    private byte[] dfName;
    private int lifeCycleStatus = -1;
    private int nettoLength = -1;
    private int totalLength = -1;

    // TODO-ASZ do not use CardException here
    public FCP(ResponseAPDU responseAPDU) throws CardException {
        ByteArrayInputStream bis = new ByteArrayInputStream(responseAPDU.getData());
        ASN1InputStream ais = new ASN1InputStream(bis);
        try {
            DERApplicationSpecific derApplicationSpecific = (DERApplicationSpecific) ais.readObject();
//            Log.d(TAG, ASN1Dump.dumpAsString(derApplicationSpecific));
            ASN1Set asn1Set = ASN1Set.getInstance(derApplicationSpecific.getObject(BERTags.SET));
            Enumeration setEnum = asn1Set.getObjects();
            while (setEnum.hasMoreElements()) {
                Object object = setEnum.nextElement();
                if(object instanceof ASN1TaggedObject) {
                    ASN1TaggedObject asn1TaggedObject = (ASN1TaggedObject) object;
                    ASN1Primitive asn1Primitive = asn1TaggedObject.getObject();
                    if (asn1Primitive instanceof ASN1OctetString) {
                        ASN1OctetString asn1OctetString = (ASN1OctetString) asn1Primitive;
                        switch (asn1TaggedObject.getTagNo()) {
                            case 0:
                                nettoLength = new BigInteger(asn1OctetString.getOctets()).intValue();
                                break;
                            case 1:
                                totalLength = new BigInteger(asn1OctetString.getOctets()).intValue();
                                break;
                            case 2:
                                fileDescriptor = asn1OctetString.getOctets();
                                break;
                            case 3:
                                fileIdentifier = asn1OctetString.getOctets();
                                break;
                            case 4:
                                dfName = asn1OctetString.getOctets();
                                break;
                            case 10:
                                lifeCycleStatus = (int) asn1OctetString.getOctets()[0];
                                if (lifeCycleStatus > 12) {
                                    lifeCycleStatus = 12;
                                } else if (lifeCycleStatus == 6) {
                                    lifeCycleStatus = 4;
                                } else if (lifeCycleStatus == 7) {
                                    lifeCycleStatus = 5;
                                }
                                break;
                        }
                    }
                }
            }
        } catch (IOException e) {
            LOG.error("io exception when reading fcp", e);
            throw new CardException("io exception when reading fcp", e);
        }
    }

    public byte[] getFileDescriptor() {
        return fileDescriptor;
    }

    public byte[] getFileIdentifier() {
        return fileIdentifier;
    }

    public byte[] getDfName() {
        return dfName;
    }

    public int getLifeCycleStatus() {
        return lifeCycleStatus;
    }

    public int getNettoLength() {
        return nettoLength;
    }

    public int getTotalLength() {
        return totalLength;
    }

    public String toString(){
        StringBuilder stringBuilder = new StringBuilder();
        if(totalLength>0){
            stringBuilder.append("total file size (byte): " + totalLength);
        }
        if(nettoLength>0){
            stringBuilder.append("\nnetto file size (byte): " + nettoLength);
        }
        if(dfName!=null){
            stringBuilder.append("\napplication identifier: " + Utilities.ByteArrayToSpacedHexString(dfName));
        }
        if(fileDescriptor!=null){
            stringBuilder.append("\nfile descriptor: " + Utilities.ByteArrayToSpacedHexString(fileDescriptor));
        }
        if(fileIdentifier!=null){
            stringBuilder.append("\nfile identifier: " + Utilities.ByteArrayToSpacedHexString(fileIdentifier));
        }
        if(lifeCycleStatus>0){
            stringBuilder.append("\nlifeCycleStatus: ");

            switch(lifeCycleStatus){
                case LCS_CREATION_STATE:
                    stringBuilder.append("CREATION_STATE");
                    break;
                case LCS_INITIALISATION_STATE:
                    stringBuilder.append("INITIALISATION_STATE");
                    break;
                case LCS_OPERATIONAL_STATE_ACTIVATED:
                    stringBuilder.append("OPERATIONAL_STATE_ACTIVATED");
                    break;
                case LCS_OPERATIONAL_STATE_DEACTIVATED:
                    stringBuilder.append("OPERATIONAL_STATE_DEACTIVATED");
                    break;
                case LCS_TERMINATION_STATE:
                    stringBuilder.append("TERMINATION_STATE");
                    break;
            }
        }
        return stringBuilder.toString().trim();
    }

}
