/*
 * HealthcardIO - a library providing functionality for accessing healthcards
 *
 * Copyright (C) 2017  gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.tr03110;

import org.spongycastle.crypto.digests.SHA1Digest;
import org.spongycastle.crypto.digests.SHA256Digest;

/**
 * @author Ronald Koenig
 */
public class KeyDerivationFunction {

    public enum Mode {

        // derive key for encryption/decryption
        ENC,

        // derive key for MAC
        MAC,

        // derive encryption keys from a password
        PASSWORD
    }

    /**
     *
     * Das MRZ-Passwort besteht aus dem SHA1-Wert der Dokumentennummer +
     * Geburtsdatum + Gültigkeitsdatum (jeweils mit Prüfziffer)
     *
     * @param documentNr
     *            Dokumentennummer plus Prüfziffer
     * @param dateOfBirth
     *            Geburtsdatum aus der MRZ plus Prüfziffer
     * @param dateOfExpiry
     *            Gültigkeitsdatum aus der MRZ plus Prüfziffer
     * @return K = SHA-1(Serial Number||Date of Birth||Date of Expiry)
     */
    public static byte[] getMRZBytes(String documentNr, String dateOfBirth,
                                     String dateOfExpiry) {
        String mrzInfo = documentNr + dateOfBirth + dateOfExpiry;
        byte[] passwordBytes = mrzInfo.getBytes();

        byte[] K = new byte[20];

        SHA1Digest sha1 = new SHA1Digest();
        sha1.update(passwordBytes, 0, passwordBytes.length);
        sha1.doFinal(K, 0);

        return K;
    }

    /**
     * derive AES-128 key
     *
     * @param K
     *            The shared secret value.
     * @param mode
     *            key derivation for ENC, MAC or derivation from password
     *
     * @return key
     */
    public static byte[] getAES128Key(byte[] K, Mode mode) {

        byte[] checksum = new byte[20];
        byte[] data = concatenateData(K, mode);
        SHA1Digest sha1 = new SHA1Digest();
        sha1.update(data, 0, data.length);
        sha1.doFinal(checksum, 0);

        //return first 16 bytes;
        byte[] key = new byte[16];
        System.arraycopy(checksum, 0, key, 0, 16);
        return key;
    }

    /**
     * derive AES-192 key
     *
     * @param K
     *            The shared secret value.
     * @param mode
     *            key derivation for ENC, MAC or derivation from password
     *
     * @return key
     */
    public static byte[] getAES192Key(byte[] K, Mode mode) {

        byte[] checksum = getAES256Key(K, mode);
        byte[] key = new byte[24];
        System.arraycopy(checksum, 0, key, 0, 24);
        return key;
    }

    /**
     * derive AES-256 key
     *
     * @param K
     *            The shared secret value.
     * @param mode
     *            key derivation for ENC, MAC or derivation from password
     *
     * @return key
     */
    public static byte[] getAES256Key(byte[] K, Mode mode) {

        byte[] checksum = new byte[32];
        byte[] data = concatenateData(K, mode);
        SHA256Digest sha256 = new SHA256Digest();
        sha256.update(data, 0, data.length);
        sha256.doFinal(checksum, 0);

        return checksum;
    }

    private static byte[] concatenateData(byte[] K, Mode mode) {
        byte[] bytes = new byte[K.length + 4];
        System.arraycopy(K, 0, bytes, 0, K.length);
        switch (mode) {
            // concatenate a 32-bit, big-endian integer
            case ENC:
                bytes[bytes.length - 1] = 1;
                break;
            case MAC:
                bytes[bytes.length - 1] = 2;
                break;
            case PASSWORD:
                bytes[bytes.length - 1] = 3;
                break;
        }
        return bytes;
    }

}
