/*
 * HealthcardIO - a library providing functionality for accessing healthcards
 *
 * Copyright (C) 2017  gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.test;

import de.gematik.security.Utilities.CardFileSystem;
import de.gematik.security.Utilities.CardItemFile;
import de.gematik.security.Utilities.EgkRSAEngine;
import de.gematik.security.Utilities.Utilities;
import de.gematik.security.crypto.CardBasedRSAPrivateKey;
import de.gematik.security.impl.HealthcardContext;
import de.gematik.security.iso7816.FCP;
import de.gematik.security.iso7816.Iso7816UnsuccessfulExecutionException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.json.JSONException;
import org.junit.Test;
import org.spongycastle.asn1.ASN1InputStream;
import org.spongycastle.asn1.ASN1OctetString;
import org.spongycastle.asn1.ASN1Sequence;
import org.spongycastle.crypto.AsymmetricBlockCipher;
import org.spongycastle.crypto.Digest;
import org.spongycastle.crypto.InvalidCipherTextException;
import org.spongycastle.crypto.digests.SHA256Digest;
import org.spongycastle.crypto.encodings.PKCS1Encoding;
import org.spongycastle.crypto.engines.RSAEngine;
import org.spongycastle.crypto.params.RSAKeyParameters;
import org.spongycastle.crypto.signers.ISO9796d2PSSSigner;
import org.spongycastle.crypto.signers.PSSSigner;

import javax.smartcardio.CardException;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.security.*;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPublicKey;
import java.security.spec.MGF1ParameterSpec;
import java.security.spec.PSSParameterSpec;
import java.util.ArrayList;
import java.util.Arrays;

import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.fail;

/**
 * @author André Szdzuy
 */
public class CryptoCommandsTest extends NfcBaseTest {
    private static final Log LOG = LogFactory.getLog(CryptoCommandsTest.class);

    @Test
    public void doRSASSA_PKCS1_V1_5_SIGN() throws CardException, InvalidCipherTextException, IOException, Iso7816UnsuccessfulExecutionException, CertificateException, JSONException {
        // RSASSA_PKCS1_V1_5_SIGN with SHA256 (see PKCS#1V2.2 chapter 8.2.1)
        LOG.info("RSASSA_PKCS1_V1_5_SIGN with SHA256");
        assertNotNull("healthcardContext is null", healthcardContext);

        byte[] input = createSignatureInput();
        byte[] signature = healthcardContext.doRSASSA_PKCS1_V1_5_SIGN(input, 0x02);

        LOG.info("RSASSA_PKCS1_V1_5 signature generated");

        verifyRSASSA_PKCS1_V1_5_SIGN(cfs, signature, input);
    }

    @Test
    public void doRSASSA_PKCS1_V1_5_SIGN_usingJCASignature() throws SignatureException, NoSuchAlgorithmException, InvalidKeyException, Iso7816UnsuccessfulExecutionException, CertificateException, CardException, JSONException, IOException, InvalidCipherTextException {
        LOG.info("useJCASignature() - RSASSA_PKCS1_V1_5_SIGN with SHA256");
        assertNotNull("healthcardContext is null", healthcardContext);

        Provider provider = Security.getProvider("HealthcardSignatureProvider");

//        Log.d(TAG, "NfcHealthCardProvider services:");
//        for (Provider.Service service:provider.getServices()) {
//            Log.d(TAG, service.getType());
//        }
        Signature signer = Signature.getInstance("SHA256withRSA", provider);
        byte[] input = createSignatureInput();
        signer.initSign(new CardBasedRSAPrivateKey(healthcardContext, 0x02));
        signer.update(input);
        byte[] signature = signer.sign();

        LOG.info("useJCASignature() - RSASSA_PKCS1_V1_5 signature generated");

        verifyRSASSA_PKCS1_V1_5_SIGN(cfs, signature, input);
    }

    @Test
    public void doPSSSigner() throws CardException, Iso7816UnsuccessfulExecutionException, CertificateException, IOException, JSONException {
        // PSSSigner
        LOG.info("PSSSigner");
        assertNotNull("healthcardContext is null", healthcardContext);

        byte[] input = createSignatureInput();
        byte[] signature = healthcardContext.doPSSSigner(input, 0x02);

        LOG.info("PSSSigner signature generated");

        verifyPSSSigner(cfs, signature, input);
    }

    @Test
    public void doPSSSigner_usingJCASignature() throws CardException, Iso7816UnsuccessfulExecutionException, CertificateException, IOException, JSONException, NoSuchAlgorithmException, InvalidKeyException, SignatureException, InvalidAlgorithmParameterException {
        // PSSSigner
        LOG.info("useJCASignature() - PSSSigner");
        assertNotNull("healthcardContext is null", healthcardContext);

        Provider provider = Security.getProvider("HealthcardSignatureProvider");
        Signature signer = Signature.getInstance("SHA256withRSAandMGF1", provider);
        PSSParameterSpec pssSpec = new PSSParameterSpec("SHA256", "MGF1", MGF1ParameterSpec.SHA256, 32, 1);
        signer.setParameter(pssSpec);
        byte[] input = createSignatureInput();
        signer.initSign(new CardBasedRSAPrivateKey(healthcardContext, 0x02));
        signer.update(input);
        byte[] signature = signer.sign();

        LOG.info("useJCASignature() - PSSSigner signature generated");

        verifyPSSSigner(cfs, signature, input);
    }

    @Test
    public void doIso9796d2PSSSigner() throws CardException, Iso7816UnsuccessfulExecutionException, CertificateException, IOException, JSONException {
        //ISO9796d2PSSSigner
        LOG.info("Iso9796d2PSSSigner");
        assertNotNull("healthcardContext is null", healthcardContext);

        byte[] input = createSignatureInput();
        byte[] signature = healthcardContext.doIso9796d2PSSSigner(input, 0x02);

        LOG.info("Iso9796d2PSSSigner signature generated");

        verifyIso9796d2PSSSigner(cfs, signature, input);
    }

    @Test
    public void doIso9796d2PSSSigner_usingJCASignature() throws CardException, Iso7816UnsuccessfulExecutionException, CertificateException, IOException, JSONException, NoSuchAlgorithmException, InvalidKeyException, SignatureException {
        //ISO9796d2PSSSigner
        LOG.info("useJCASignature() - Iso9796d2PSSSigner");
        assertNotNull("healthcardContext is null", healthcardContext);

        Provider provider = Security.getProvider("HealthcardSignatureProvider");
        Signature signer = Signature.getInstance("SHA256withIso9796d2PSS", provider);
        byte[] input = createSignatureInput();
        signer.initSign(new CardBasedRSAPrivateKey(healthcardContext, 0x02));
        signer.update(input);
        byte[] signature = signer.sign();

        LOG.info("useJCASignature() - Iso9796d2PSSSigner signature generated");

        verifyIso9796d2PSSSigner(cfs, signature, input);
    }

    private void verifyRSASSA_PKCS1_V1_5_SIGN(CardFileSystem cfs, byte[] signature, byte[] input) throws Iso7816UnsuccessfulExecutionException, CertificateException, CardException, JSONException, IOException, InvalidCipherTextException {
        X509Certificate x509Cert = getX509Cert(cfs, healthcardContext);
        RSAPublicKey rsaPublicKey = (RSAPublicKey) x509Cert.getPublicKey();
        RSAKeyParameters rsaPublicKeyParameters = new RSAKeyParameters(false, rsaPublicKey.getModulus(), rsaPublicKey.getPublicExponent());

        //verify signature
        Digest digest = new SHA256Digest();
        byte[] hRec = new byte[digest.getDigestSize()];
        byte[] hCalc = new byte[digest.getDigestSize()];

        //decode signature
        AsymmetricBlockCipher cipher = new PKCS1Encoding(new RSAEngine());
        cipher.init(false, rsaPublicKeyParameters);
        byte[] digestInfoReceived = cipher.processBlock(signature, 0, signature.length);

        //get hash
        ByteArrayInputStream bis = new ByteArrayInputStream(digestInfoReceived);
        ASN1InputStream ais = new ASN1InputStream(bis);
        ASN1Sequence asn1Sequence = (ASN1Sequence) ais.readObject();
        hRec = ((ASN1OctetString) asn1Sequence.getObjectAt(1)).getOctets();

        //calc hash
        digest.update(input, 0, input.length);
        digest.doFinal(hCalc, 0);

        //compare
        if (Arrays.equals(hRec, hCalc)) {
            LOG.info("RSASSA_PKCS1_V1_5 signature successfully verified");
        } else {
            fail("RSASSA_PKCS1_V1_5 signature verification failed");
        }
    }

    private void verifyPSSSigner(CardFileSystem cfs, byte[] signature, byte[] input) throws Iso7816UnsuccessfulExecutionException, CertificateException, CardException, JSONException, IOException {
        // Verify signature
        X509Certificate x509Cert = getX509Cert(cfs, healthcardContext);
        RSAPublicKey rsaPublicKey = (RSAPublicKey) x509Cert.getPublicKey();
        RSAKeyParameters rsaPublicKeyParameters = new RSAKeyParameters(false, rsaPublicKey.getModulus(), rsaPublicKey.getPublicExponent());
        PSSSigner PSSSigner = new PSSSigner(new RSAEngine(), new SHA256Digest(), 32);
        PSSSigner.init(false, rsaPublicKeyParameters);
        PSSSigner.update(input, 0, input.length);
        if (PSSSigner.verifySignature(signature)) {
            LOG.info("PSSSigner signature successfully verified");
        } else {
            fail("PSSSigner signature verification failed");
        }
    }

    private void verifyIso9796d2PSSSigner(CardFileSystem cfs, byte[] signature, byte[] input) throws Iso7816UnsuccessfulExecutionException, CertificateException, CardException, JSONException, IOException {
        // Verify signature
        X509Certificate x509Cert = getX509Cert(cfs, healthcardContext);
        RSAPublicKey rsaPublicKey = (RSAPublicKey) x509Cert.getPublicKey();
        RSAKeyParameters rsaPublicKeyParameters = new RSAKeyParameters(false, rsaPublicKey.getModulus(), rsaPublicKey.getPublicExponent());
        ISO9796d2PSSSigner iso9796d2PSSSigner = new ISO9796d2PSSSigner(new EgkRSAEngine(), new SHA256Digest(), 32, true);

        iso9796d2PSSSigner.init(false, rsaPublicKeyParameters);
        iso9796d2PSSSigner.update(input, 0, input.length);
        if (iso9796d2PSSSigner.verifySignature(signature)) {
            LOG.info("Iso9796d2PSSSigner signature successfully verified");
        } else {
            LOG.info("Iso9796d2PSSSigner signature verification failed");
        }

        String recovered = null;
        if (iso9796d2PSSSigner.getRecoveredMessage() != null) {
            recovered = Utilities.ByteArrayToSpacedHexString(iso9796d2PSSSigner.getRecoveredMessage());
        }
        String m = (iso9796d2PSSSigner.hasFullMessage() ? "complete" : "incomplete") + " message recovered: " + recovered;
        LOG.info(m);
    }

    private X509Certificate getX509Cert(CardFileSystem cfs, HealthcardContext channel) throws Iso7816UnsuccessfulExecutionException, CardException, CertificateException, IOException, JSONException {
        FCP parent = channel.selectADFAndReturnFCP("A000000167455349474E");  // DF.ESIGN
        CardItemFile cardItemFile = channel.readEF(cfs, parent, "C5 00"); // EF.AUT
        ArrayList<byte[]> data = cardItemFile.getData();

        X509Certificate x509Cert = Utilities.getX509Certificate(data.get(0));
        assertNotNull(x509Cert);
        return x509Cert;
    }

    private byte[] createSignatureInput() {
        byte[] input = new byte[256];
        for (int i = 0; i < input.length; i++) {
            input[i] = (byte) i;
        }
        return input;
    }

}
